// Frame rendering with SVP shaders, based on MVFlowFps.cpp code
// Copyright(c)2011-2012 Chainik @forum.svp-team.com

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MSmoothFps.h"
#include "MVFinest.h"

#include "MaskFun.h"
#include "commonfunctions.h"

#include <math.h>
#include <string>

MSmoothFps::MSmoothFps(PClip _child, PClip super, PClip _mvbw, PClip _mvfw, PClip _finest,
                unsigned int _num, unsigned int _den,
                int _algo, int _ml,
                bool _block, bool _blend,
                int _sadml,double _sadgamma,
                int nSCD1, int nSCD2,
                int _flags, int gpuId,
                IScriptEnvironment* env) :
    GenericVideoFilter(_child),
    MVFilter(_mvfw, "MSmoothFps", env),
    mvClipB(_mvbw, nSCD1, nSCD2, env),
    mvClipF(_mvfw, nSCD1, nSCD2, env)
{
// FPS calculation magic
    unsigned int numeratorOld = vi.fps_numerator;
    unsigned int denominatorOld = vi.fps_denominator;
    unsigned int numerator;
    unsigned int denominator;

    if (_num != 0 && _den != 0)
    {
        numerator = _num;
        denominator = _den;
    }
    else if (numeratorOld < (1<<30))
    {
        numerator = (numeratorOld<<1); // double fps by default
        denominator = denominatorOld;
    }
    else // very big numerator
    {
        numerator = numeratorOld;
        denominator = (denominatorOld>>1);// double fps by default
    }

    //  safe for big numbers since v2.1
    fa = __int64(denominator)*__int64(numeratorOld);
    fb = __int64(numerator)*__int64(denominatorOld);
    __int64 fgcd = gcd(fa, fb); // general common divisor
    fa /= fgcd;
    fb /= fgcd;

    vi.SetFPS(numerator, denominator);
    vi.num_frames = (int)(1 + __int64(vi.num_frames-1) * fb/fa );

// check vector clips
    if (!mvClipB.IsBackward())
        env->ThrowError("MSmoothFps: wrong backward vectors");
    if (mvClipF.IsBackward())
        env->ThrowError("MSmoothFps: wrong forward vectors");

    CheckSimilarity(mvClipB, "mvbw", env);
    CheckSimilarity(mvClipF, "mvfw", env);

// check super clip params
    if (nWidth  != vi.width || nHeight  != vi.height )
        env->ThrowError("MSmoothFps: inconsistent source and vector frame size");

    if (!((nWidth + nHPadding*2) == super->GetVideoInfo().width && (nHeight + nVPadding*2) <= super->GetVideoInfo().height))
        env->ThrowError("MSmoothFps: inconsistent clips frame size!");

    // get parameters of prepared super clip - v2.0
    SuperParams64Bits params;
    memcpy(&params, &super->GetVideoInfo().num_audio_samples, 8);
    int nHeightS = params.nHeight;
    int nSuperHPad = params.nHPad;
    int nSuperWidth = super->GetVideoInfo().width;

    if (nHeight != nHeightS || nWidth != nSuperWidth-nSuperHPad*2)
        env->ThrowError("MSmoothFps : wrong super frame clip");

// determine blocks count, padding, etc.
    // may be padded for full frame cover
    nBlkXP = (nBlkX*(nBlkSizeX - nOverlapX) + nOverlapX < nWidth) ? nBlkX+1 : nBlkX;
    nBlkYP = (nBlkY*(nBlkSizeY - nOverlapY) + nOverlapY < nHeight) ? nBlkY+1 : nBlkY;
    nWidthP = nBlkXP*(nBlkSizeX - nOverlapX) + nOverlapX;
    nHeightP = nBlkYP*(nBlkSizeY - nOverlapY) + nOverlapY;
    // for YV12
    nWidthPUV = nWidthP/2;
    nHeightPUV = nHeightP/yRatioUV;
    nWidthUV = nWidth/2;
    nHeightUV = nHeight/yRatioUV;

    nHPaddingUV = nHPadding/2;
    nVPaddingUV = nVPadding/yRatioUV;

    VPitchY = (nWidthP + 15) & (~15);
    VPitchUV = (nWidthPUV + 15) & (~15);

// init processing params
    algo=_algo;
    switch(algo)
    {
    case AlgoFast_1dir:
    case AlgoFast_2dir:
    case AlgoNoMask1:
    case AlgoNoMask2:
    case AlgoNoMask3:
    case AlgoSimple:
    case AlgoNormal:
    case AlgoExtended: break;

    default: env->ThrowError("MSmoothFps: incorrect 'algo' value.");
    }

    flags=_flags;

    occScaleFactor = (_ml<=0 ? 1000000 : _ml);
    blendOnSceneChange=_blend;
    flowBlocks=_block;

    applySADMask=(_sadml>0);
    sadNormFactor=(applySADMask ? 1.0/_sadml : 1.0);
    sadGamma=_sadgamma;

    if(_ml<=0 && algo>=AlgoSimple && algo<=AlgoExtended)
        algo=AlgoNoMask1;

// load libflowXXX.dll
    fnFlowLib_GetVersion pVersion;

    const char *flowLibPath=0;
    try
    { flowLibPath=env->GetVar("svp_flow_lib").AsString(); }
    catch(...)
    { env->ThrowError("MSmoothFps: svp_flow_lib must be declared"); }

    HINSTANCE hLib=LoadLibrary(flowLibPath);
    if(hLib)
    {
        pFlowLib_Init=(fnFlowLib_Init)GetProcAddress(hLib,"flowLib_Init");
        pFlowLib_Clear=(fnFlowLib_Clear)GetProcAddress(hLib,"flowLib_Clear");
        pFlowLib_SetTime=(fnFlowLib_SetTime)GetProcAddress(hLib,"flowLib_SetTime");
        pFlowLib_Calculate=(fnFlowLib_Calculate)GetProcAddress(hLib,"flowLib_Calculate");

        pVersion=(fnFlowLib_GetVersion)GetProcAddress(hLib,"flowLib_GetVersion");
    }
    else
        env->ThrowError("MSmoothFps: unable to load library given in svp_flow_lib");

    if(!pFlowLib_Init || !pFlowLib_Clear || !pFlowLib_SetTime ||
            !pFlowLib_Calculate || !pVersion)
        env->ThrowError("MSmoothFps: invalid FlowLib given in svp_flow_lib");

    flowLibVersion=pVersion();

    if(flowLibVersion<FLOWLIB_VERSION)
        env->ThrowError("MSmoothFps: FlowLib version is too old");

    int flowLibFlags=0;
    if(flowBlocks) flowLibFlags|=FLOWLIB_FLAG_BLOCK;
    if(applySADMask) flowLibFlags|=FLOWLIB_FLAG_SAD;

    if(flowLibVersion&FLOWLIB_TYPE_GPU)
    {
        int gpu=((gpuId/10)<<16) + (gpuId%10);
        flowLibFlags|=(gpu<<8);
    }

    hFlowLib=pFlowLib_Init(nWidth, nHeight,
                           nBlkSizeX-nOverlapX,nBlkSizeY-nOverlapY,
                           nBlkXP,nBlkYP,
                           nBlkSizeX/2,nBlkSizeY/2,
                           yRatioUV,nPel,
                           algo,flowLibFlags);
    if(!hFlowLib)
        env->ThrowError("MSmoothFps: unable to init FlowLib.");

    ignorePel=(flowLibVersion&FLOWLIB_TYPE_GPU || _finest==0);// || flowBlocks);
    finest=(ignorePel ? super : _finest);

// init work arrays
    int sz=nBlkXP*nBlkYP;
    VX_B = new BYTE[sz];
    VY_B = new BYTE[sz];
    VX_F = new BYTE[sz];
    VY_F = new BYTE[sz];
    VX_BB = new BYTE[sz];
    VY_BB = new BYTE[sz];
    VX_FF = new BYTE[sz];
    VY_FF = new BYTE[sz];
    OccMask_B = new BYTE[sz];
    OccMask_F = new BYTE[sz];
    SADMask_B = (applySADMask ? new BYTE[sz] : 0);
    SADMask_F = (applySADMask ? new BYTE[sz] : 0);

    if ( (pixelType & VideoInfo::CS_YUY2) == VideoInfo::CS_YUY2)
        DstPlanes =  new YUY2Planes(nWidth, nHeight);
}

MSmoothFps::~MSmoothFps()
{
    pFlowLib_Clear(hFlowLib);

    delete[] VX_B;
    delete[] VY_B;
    delete[] VX_F;
    delete[] VY_F;
    delete[] VX_BB;
    delete[] VY_BB;
    delete[] VX_FF;
    delete[] VY_FF;

    delete[] OccMask_B;
    delete[] OccMask_F;

    if(applySADMask)
    {
        delete[] SADMask_B;
        delete[] SADMask_F;
    }

    if((pixelType & VideoInfo::CS_YUY2) == VideoInfo::CS_YUY2)
        delete DstPlanes;
}

//from MVMask
BYTE MSmoothFps::calcSAD(unsigned int s)
{
	double l = 255 * pow((s*4*sadNormFactor)/(nBlkSizeX*nBlkSizeY), sadGamma);
	return (BYTE)min(l,255.0);
}

//-------------------------------------------------------------------------
PVideoFrame __stdcall MSmoothFps::GetFrame(int n, IScriptEnvironment* env)
{
    int nleft = (int) ( __int64(n)* fa/fb );
    int time256 = int( (double(n)*double(fa)/double(fb) - nleft)*256 + 0.5);

    BYTE *pDst[3];
    const BYTE *pRef[3], *pSrc[3];
    int nDstPitches[3], nRefPitches[3], nSrcPitches[3];
    unsigned char *pDstYUY2;
    int nDstPitchYUY2;

    int off = mvClipB.GetDeltaFrame(); // integer offset of reference frame
    // usually off must be = 1
    if (off > 1)
        time256 = time256/off;
    int nright = nleft+off;

    if(time256==0)
        return child->GetFrame(nleft,env);

    if(time256==256)
        return child->GetFrame(nright,env);

    bool use_vector_F=((algo==AlgoFast_2dir && time256<128) || algo>=AlgoNoMask1);
    bool use_vector_B=(algo!=AlgoFast_2dir || time256>=128);
    bool inverseTime=(use_vector_B && !use_vector_F);

    if(inverseTime)
        time256=256-time256;

    if(use_vector_F)
        mvClipF.Update(mvClipF.GetFrame(nright, env), env);// forward from current to next

    if(use_vector_B)
        mvClipB.Update(mvClipB.GetFrame(nleft, env), env);// backward from next to current

    if((!use_vector_B || mvClipB.IsUsable()) && (!use_vector_F || mvClipF.IsUsable()))
    {
        PVideoFrame dst = env->NewVideoFrame(vi);
        PVideoFrame	src	= finest->GetFrame(nleft, env);
        PVideoFrame ref = finest->GetFrame(nright, env);

        if(applySADMask)
        {
            for(int x=0;x<nBlkXP;x++)
                for(int y=0;y<nBlkYP;y++)
                {
                    int j1=y*nBlkXP+x;
                    int j2=y*nBlkX+x;

                    if(x<nBlkX && y<nBlkY)
                    {
                        SADMask_F[j1] = calcSAD(mvClipF.GetBlock(0,j2).GetSAD());
                        SADMask_B[j1] = calcSAD(mvClipB.GetBlock(0,j2).GetSAD());
                    }
                    else
                    {
                        SADMask_F[j1] = SADMask_F[j1-(x>=nBlkX ? 1 : nBlkXP)];
                        SADMask_B[j1] = SADMask_B[j1-(x>=nBlkX ? 1 : nBlkXP)];
                    }
                }
        }

        pFlowLib_SetTime(hFlowLib,time256);

        if((pixelType & VideoInfo::CS_YUY2) == VideoInfo::CS_YUY2)
        {
            // planar data packed to interleaved format (same as interleved2planar by kassandro) - v2.0.0.5
            pSrc[0] = src->GetReadPtr();
            pSrc[1] = pSrc[0] + src->GetRowSize()/2;
            pSrc[2] = pSrc[1] + src->GetRowSize()/4;
            nSrcPitches[0] = src->GetPitch();
            nSrcPitches[1] = nSrcPitches[0];
            nSrcPitches[2] = nSrcPitches[0];

            pRef[0] = ref->GetReadPtr();
            pRef[1] = pRef[0] + ref->GetRowSize()/2;
            pRef[2] = pRef[1] + ref->GetRowSize()/4;
            nRefPitches[0] = ref->GetPitch();
            nRefPitches[1] = nRefPitches[0];
            nRefPitches[2] = nRefPitches[0];

            pDstYUY2 = dst->GetWritePtr();
            nDstPitchYUY2 = dst->GetPitch();
            pDst[0] = DstPlanes->GetPtr();
            pDst[1] = DstPlanes->GetPtrU();
            pDst[2] = DstPlanes->GetPtrV();
            nDstPitches[0]  = DstPlanes->GetPitch();
            nDstPitches[1]  = DstPlanes->GetPitchUV();
            nDstPitches[2]  = DstPlanes->GetPitchUV();
        }
        else
        {
            pDst[0] = YWPLAN(dst);
            pDst[1] = UWPLAN(dst);
            pDst[2] = VWPLAN(dst);
            nDstPitches[0] = YPITCH(dst);
            nDstPitches[1] = UPITCH(dst);
            nDstPitches[2] = VPITCH(dst);

            pRef[0] = YRPLAN(ref);
            pRef[1] = URPLAN(ref);
            pRef[2] = VRPLAN(ref);
            nRefPitches[0] = YPITCH(ref);
            nRefPitches[1] = UPITCH(ref);
            nRefPitches[2] = VPITCH(ref);

            pSrc[0] = YRPLAN(src);
            pSrc[1] = URPLAN(src);
            pSrc[2] = VRPLAN(src);
            nSrcPitches[0] = YPITCH(src);
            nSrcPitches[1] = UPITCH(src);
            nSrcPitches[2] = VPITCH(src);
        }

        int nOffsetY = nRefPitches[0] * nVPadding + nHPadding;
        int nOffsetUV = nRefPitches[1] * nVPaddingUV + nHPaddingUV;

        if(!ignorePel)
        {
            nOffsetY*=nPel;
            nOffsetUV*=nPel;
        }

        if(use_vector_B)
        {
            MakeVectorSmallMasks(mvClipB, nBlkX, nBlkY, VX_B, nBlkXP, VY_B, nBlkXP);
            if (nBlkXP > nBlkX) // fill right
            {
                for (int j=0; j<nBlkY; j++)
                {
                    VX_B[j*nBlkXP + nBlkX] = min(VX_B[j*nBlkXP + nBlkX-1],128);
                    VY_B[j*nBlkXP + nBlkX] = VY_B[j*nBlkXP + nBlkX-1];
                }
            }
            if (nBlkYP > nBlkY) // fill bottom
            {
                for (int i=0; i<nBlkXP; i++)
                {
                    VX_B[nBlkXP*nBlkY +i] = VX_B[nBlkXP*(nBlkY-1) +i];
                    VY_B[nBlkXP*nBlkY +i] = min(VY_B[nBlkXP*(nBlkY-1) +i],128);
                }
            }
        }
        if(algo>=AlgoSimple) //need occlusion masks
        {
            // analyse vectors field to detect occlusion
            MakeVectorOcclusionMaskTime(mvClipB, nBlkX, nBlkY, occScaleFactor, 1.0, nPel, OccMask_B, nBlkXP, (256-time256), nBlkSizeX - nOverlapX, nBlkSizeY - nOverlapY);
            if (nBlkXP > nBlkX) // fill right
                for (int j=0; j<nBlkY; j++)
                    OccMask_B[j*nBlkXP + nBlkX] = OccMask_B[j*nBlkXP + nBlkX-1];
            if (nBlkYP > nBlkY) // fill bottom
                for (int i=0; i<nBlkXP; i++)
                    OccMask_B[nBlkXP*nBlkY +i] = OccMask_B[nBlkXP*(nBlkY-1) +i];
        }

        if(use_vector_F)
        {
            MakeVectorSmallMasks(mvClipF, nBlkX, nBlkY, VX_F, nBlkXP, VY_F, nBlkXP);
            if (nBlkXP > nBlkX) // fill right
            {
                for (int j=0; j<nBlkY; j++)
                {
                    VX_F[j*nBlkXP + nBlkX] = min(VX_F[j*nBlkXP + nBlkX-1],128);
                    VY_F[j*nBlkXP + nBlkX] = VY_F[j*nBlkXP + nBlkX-1];
                }
            }
            if (nBlkYP > nBlkY) // fill bottom
            {
                for (int i=0; i<nBlkXP; i++)
                {
                    VX_F[nBlkXP*nBlkY +i] = VX_F[nBlkXP*(nBlkY-1) +i];
                    VY_F[nBlkXP*nBlkY +i] = min(VY_F[nBlkXP*(nBlkY-1) +i],128);
                }
            }
        }
        if(algo>=AlgoSimple) //need occlusion masks
        {
            MakeVectorOcclusionMaskTime(mvClipF, nBlkX, nBlkY, occScaleFactor, 1.0, nPel, OccMask_F, nBlkXP, time256, nBlkSizeX - nOverlapX, nBlkSizeY - nOverlapY);
            if (nBlkXP > nBlkX) // fill right
                for (int j=0; j<nBlkY; j++)
                    OccMask_F[j*nBlkXP + nBlkX] = OccMask_F[j*nBlkXP + nBlkX-1];
            if (nBlkYP > nBlkY) // fill bottom
                for (int i=0; i<nBlkXP; i++)
                    OccMask_F[nBlkXP*nBlkY +i] = OccMask_F[nBlkXP*(nBlkY-1) +i];
        }

        bool extraVecsReady=false;

        if(algo==AlgoExtended || algo>=90)  // slow method with extra frames
        {
            // Get motion info from more frames for occlusion areas
            mvClipF.Update(mvClipF.GetFrame(nleft, env), env);// forward from prev to cur
            mvClipB.Update(mvClipB.GetFrame(nright, env), env);// backward from next next to next

            if(mvClipB.IsUsable() && mvClipF.IsUsable())
            {
                // get vector mask from extra frames
                MakeVectorSmallMasks(mvClipB, nBlkX, nBlkY, VX_BB, nBlkXP, VY_BB, nBlkXP);
                MakeVectorSmallMasks(mvClipF, nBlkX, nBlkY, VX_FF, nBlkXP, VY_FF, nBlkXP);
                if (nBlkXP > nBlkX) // fill right
                {
                    for (int j=0; j<nBlkY; j++)
                    {
                        VX_BB[j*nBlkXP + nBlkX] = min(VX_BB[j*nBlkXP + nBlkX-1],128);
                        VY_BB[j*nBlkXP + nBlkX] = VY_BB[j*nBlkXP + nBlkX-1];
                        VX_FF[j*nBlkXP + nBlkX] = min(VX_FF[j*nBlkXP + nBlkX-1],128);
                        VY_FF[j*nBlkXP + nBlkX] = VY_FF[j*nBlkXP + nBlkX-1];
                    }
                }
                if (nBlkYP > nBlkY) // fill bottom
                {
                    for (int i=0; i<nBlkXP; i++)
                    {
                        VX_BB[nBlkXP*nBlkY +i] = VX_BB[nBlkXP*(nBlkY-1) +i];
                        VY_BB[nBlkXP*nBlkY +i] = min(VY_BB[nBlkXP*(nBlkY-1) +i],128);
                        VX_FF[nBlkXP*nBlkY +i] = VX_FF[nBlkXP*(nBlkY-1) +i];
                        VY_FF[nBlkXP*nBlkY +i] = min(VY_FF[nBlkXP*(nBlkY-1) +i],128);
                    }
                }

                extraVecsReady=true;
            }
        }

        int flow_algo;
        switch(algo)
        {
        case AlgoExtended:
            flow_algo=(extraVecsReady ? AlgoExtended : AlgoNormal);
            break;
        case AlgoFast_2dir:
            flow_algo=(inverseTime ? AlgoFast_1dir : AlgoFast_2dir);
            break;

        default: flow_algo=algo;
        }

        if(flowBlocks)
            flow_algo|=AlgoProcessBlocks;

        pFlowLib_Calculate(hFlowLib,flow_algo,
                           pDst[0], nDstPitches[0],
                           pDst[1], pDst[2], nDstPitches[1],
                           nright, pRef[0] + nOffsetY, nleft, pSrc[0] + nOffsetY, nRefPitches[0],
                           pRef[1] + nOffsetUV, pSrc[1] + nOffsetUV,
                           pRef[2] + nOffsetUV, pSrc[2] + nOffsetUV, nRefPitches[1],
                           VX_B, VX_F, VY_B, VY_F,
                           VX_BB, VX_FF, VY_BB, VY_FF,
                           OccMask_B, OccMask_F,
                           SADMask_B, SADMask_F);

        if((pixelType & VideoInfo::CS_YUY2) == VideoInfo::CS_YUY2)
            YUY2FromPlanes(pDstYUY2, nDstPitchYUY2, nWidth, nHeight,
                           pDst[0], nDstPitches[0], pDst[1], pDst[2], nDstPitches[1], true);

        return dst;
    }
    else //scene change
    {
        PVideoFrame src = child->GetFrame(nleft,env);
        PVideoFrame ref = child->GetFrame(nright,env);

        if(inverseTime)
            time256=256-time256;

        if(!blendOnSceneChange)
            return (time256<128 ? src : ref);
        else //let's blend src with ref frames like ConvertFPS
        {
            PVideoFrame dst=env->NewVideoFrame(vi);

            if ( (pixelType & VideoInfo::CS_YUY2) == VideoInfo::CS_YUY2 )
            {
                pSrc[0] = src->GetReadPtr(); // we can blend YUY2
                nSrcPitches[0] = src->GetPitch();
                pRef[0] = ref->GetReadPtr();
                nRefPitches[0]  = ref->GetPitch();
                pDstYUY2 = dst->GetWritePtr();
                nDstPitchYUY2 = dst->GetPitch();
                Blend(pDstYUY2, pSrc[0], pRef[0], nHeight, nWidth*2, nDstPitchYUY2, nSrcPitches[0], nRefPitches[0], time256, true);
            }
            else
            {
                pDst[0] = YWPLAN(dst);
                pDst[1] = UWPLAN(dst);
                pDst[2] = VWPLAN(dst);
                nDstPitches[0] = YPITCH(dst);
                nDstPitches[1] = UPITCH(dst);
                nDstPitches[2] = VPITCH(dst);

                pRef[0] = YRPLAN(ref);
                pRef[1] = URPLAN(ref);
                pRef[2] = VRPLAN(ref);
                nRefPitches[0] = YPITCH(ref);
                nRefPitches[1] = UPITCH(ref);
                nRefPitches[2] = VPITCH(ref);

                pSrc[0] = YRPLAN(src);
                pSrc[1] = URPLAN(src);
                pSrc[2] = VRPLAN(src);
                nSrcPitches[0] = YPITCH(src);
                nSrcPitches[1] = UPITCH(src);
                nSrcPitches[2] = VPITCH(src);
                // blend with time weight
                Blend(pDst[0], pSrc[0], pRef[0], nHeight, nWidth, nDstPitches[0], nSrcPitches[0], nRefPitches[0], time256, true);
                Blend(pDst[1], pSrc[1], pRef[1], nHeightUV, nWidthUV, nDstPitches[1], nSrcPitches[1], nRefPitches[1], time256, true);
                Blend(pDst[2], pSrc[2], pRef[2], nHeightUV, nWidthUV, nDstPitches[2], nSrcPitches[2], nRefPitches[2], time256, true);
            }

            return dst;
        }
    }
}
